/*
   PlayStation Net-Yarouze! development tool
   SIOCONS for UNIX
   1996 (c)Go Watanabe   All rights Reserved.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "siocons.h"

#ifndef DEVICENAME
#define DEVICENAME "/dev/cua00"
#endif

#ifndef BAUDRATE
#define BAUDRATE	115200
#endif

static char    *device = DEVICENAME;
static int      speed = BAUDRATE;

int             ps_fd;		/* PlayStation's port */

static          speed_t
speed_select(int s)
{
	speed_t         ret;
	if (s == 9600)
		ret = B9600;
	else if (s == 19200)
		ret = B19200;
	else if (s == 38400)
		ret = B38400;
#if 0
#ifdef B57600
	else if (s == 76800)
		ret = B76800;
#endif
#endif
#ifdef B115200
	else if (s == 115200)
		ret = B115200;
#endif
	else
		ret = B9600;

	return ret;
}

static int      save_fd = -1;
static struct termios save_termios;

static void
tty_atexit(void)
{
	if (save_fd >= 0) {
		tcsetattr(save_fd, TCSAFLUSH, &save_termios);
		save_fd = -1;
	}
}

/* change tty to raw mode */
int
tty_raw(int fd)
{
	struct termios  buf;

	if (tcgetattr(fd, &save_termios) < 0)
		return -1;
	buf = save_termios;
	buf.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	buf.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	buf.c_oflag &= ~(OPOST);
	buf.c_cflag &= ~(CSIZE | PARENB);
	buf.c_cflag |= (CS8);
	buf.c_cc[VMIN] = 1;
	buf.c_cc[VTIME] = 0;
	if (tcsetattr(fd, TCSAFLUSH, &buf) < 0)
		return -1;
	save_fd = fd;
	return 0;
}

int
tty_nosig_noeof(int fd)
{
	struct termios  buf;

	buf = save_termios;
	buf.c_lflag &= ~(ISIG);
	buf.c_iflag &= ~(BRKINT);
	buf.c_cc[VMIN] = 1;
	buf.c_cc[VTIME] = 0;
	buf.c_cc[VEOF] = _POSIX_VDISABLE;

	if (tcsetattr(fd, TCSAFLUSH, &buf) < 0)
		return -1;
	return 0;
}

/* open tty on raw mode */
int
tty_open_raw(char *dev, speed_t s)
{
	struct termios  buf;
	int             fd;

	if ((fd = open(dev, O_RDWR)) < 0)
		return -1;

	if (tcgetattr(fd, &save_termios) < 0)
		return -1;

	/* to raw mode */
	buf = save_termios;
	buf.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	buf.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON | IXOFF | IXON);
	buf.c_oflag &= ~(OPOST);
	buf.c_cflag &= ~(CSIZE | PARENB);
	buf.c_cflag |= (CS8 | CLOCAL);
	buf.c_cc[VMIN] = 1;
	buf.c_cc[VTIME] = 0;

	cfsetospeed(&buf, s);
	cfsetispeed(&buf, s);

	if (tcsetattr(fd, TCSAFLUSH, &buf) < 0)
		return -1;

	return fd;
}

#define BUFFSIZE 1

static void
loop(void)
{
	int             n;
	u_char         *buf;
	fd_set          rset;

	if ((buf = malloc(BUFFSIZE)) == NULL) {
		perror("malloc");
		exit(1);
	}
	FD_ZERO(&rset);
	for (;;) {
		FD_SET(ps_fd, &rset);
		FD_SET(STDIN_FILENO, &rset);

		if (select(ps_fd + 1, &rset, NULL, NULL, NULL) < 0) {
			perror("select");
			exit(1);
		}
		/* from PlayStation */
		if (FD_ISSET(ps_fd, &rset)) {
			
			if ((n = read(ps_fd, buf, BUFFSIZE)) < 0) {
				perror("read");
				exit(1);
			}
			/* echo back */
			if (write(STDOUT_FILENO, buf, n) != n) {
				perror("write");
				exit(1);
			}
		}
		/* from STDIN */
		if (FD_ISSET(STDIN_FILENO, &rset)) {

			if ((n = read(STDIN_FILENO, buf, BUFFSIZE)) <= 0)
				break;

			/* escape code */
			if (buf[0] == '~') {

				/* clear input */
				{
					char            c = 003;	/* ^C */
					if (write(ps_fd, &c, 1) < 0) {
						perror("write");
						exit(1);
					}
					if ((n = read(ps_fd, buf, BUFFSIZE)) < 0) {
						perror("read");
						exit(1);
					}
				}

				tty_atexit();
				if (tty_nosig_noeof(STDIN_FILENO) < 0) {
					perror("tty_nosig");
					exit(1);
				}

				siocons_local();

				if (tty_raw(STDIN_FILENO) < 0) {
					perror("tty_raw");
					exit(1);
				}
			} else {
				/* send to PlayStation */
				if (write(ps_fd, buf, n) != n) {
					perror("write");
					exit(1);
				}
			}
		}
	}
}

int
main(int argc, char *argv[])
{
	int c;
	char *usage =
		"uasge: %s [-B baudrate] [-d device] [autofile]\n";

	opterr = 0;
	while ((c = getopt(argc, argv, "B:d:")) != EOF) {
		switch (c) {
		case 'B':
			speed = atoi(optarg);
			break;
		case 'd':
			device = optarg;
			break;
		default:
			fprintf(stderr, usage, argv[0]);
			exit(1);
		}
	}

	if (!isatty(STDIN_FILENO)) {
		fprintf(stderr, "STDIN is not interactive.\n");
		exit(1);
	}

	if ((ps_fd = tty_open_raw(device, speed_select(speed))) < 0) {
		fprintf(stderr, "open \"%s\": %s\n", device, strerror(errno));
		exit(1);
	}
	atexit(tty_atexit);

	/* auto file */
	if (argc > 1) {

		FILE *fp;
		if ((fp = fopen(argv[1],"r")) == NULL) {
			perror(argv[1]);
			exit(1);
		}

		if (tty_nosig_noeof(STDIN_FILENO) < 0) {
			perror("tty_nosig");
			exit(1);
		}
		siocons_autofile(fp);
		fclose(fp);
		tty_atexit();
	}
	
	if (tty_raw(STDIN_FILENO) < 0) {
		perror("tty_raw");
		exit(1);
	}
	loop();

	exit(0);
}
